@@ -21,7 +21,7 @@ from igneous.tasks import (
   ContrastNormalizationTask, DownsampleTask,
   TransferTask, TouchTask, LuminanceLevelsTask,
   HyperSquareConsensusTask, # HyperSquareTask,
-  ImageShardTransferTask
+  ImageShardTransferTask, ImageShardDownsampleTask
 )
 from igneous.shards import image_shard_shape_from_spec
 from igneous.types import ShapeType
@@ -35,6 +35,7 @@ __all__  = [
   "create_blackout_tasks",
   "create_touch_tasks",
   "create_downsampling_tasks", 
+  "create_image_shard_downsample_tasks",
   "create_deletion_tasks",
   "create_transfer_tasks",
   "create_image_shard_transfer_tasks",
@@ -46,6 +47,10 @@ __all__  = [
   "create_contrast_normalization_tasks",
 ]
 
+# A reasonable size for processing large
+# image chunks.
+MEMORY_TARGET = int(3.5e9)
+
 def create_blackout_tasks(
     cloudpath, bounds, 
     mip=0, shape=(2048, 2048, 64), 
@@ -115,7 +120,7 @@ def create_touch_tasks(
   class TouchTaskIterator(FinelyDividedTaskIterator):
     def task(self, shape, offset):
       bounded_shape = min2(shape, vol.bounds.maxpt - offset)
-      return igneous.tasks.TouchTask(
+      return partial(igneous.tasks.TouchTask,
         cloudpath=cloudpath,
         shape=bounded_shape.clone(),
         offset=offset.clone(),
@@ -252,7 +257,7 @@ def create_sharded_image_info(
   chunk_size: ShapeType,
   encoding: str,
   dtype: Any,
-  uncompressed_shard_bytesize: int = 3.5e9, 
+  uncompressed_shard_bytesize: int = MEMORY_TARGET, 
   max_shard_index_bytes: int = 8192, # 2^13
   max_minishard_index_bytes: int = 40000,
   max_labels_per_minishard: int = 4000
@@ -391,7 +396,7 @@ def create_image_shard_transfer_tasks(
   dest_voxel_offset: Optional[ShapeType] = None,
   agglomerate: bool = False, 
   timestamp: bool = None,
-  memory_target: int = 3.5e9,
+  memory_target: int = MEMORY_TARGET,
 ):
   src_vol = CloudVolume(src_layer_path, mip=mip)
 
@@ -482,6 +487,79 @@ def create_image_shard_transfer_tasks(
 
   return ImageShardTransferTaskIterator(bounds, shape)
 
+def create_image_shard_downsample_tasks(
+  cloudpath, mip=0, fill_missing=False, 
+  sparse=False, chunk_size=None,
+  encoding=None, memory_target=MEMORY_TARGET,
+  agglomerate=False, timestamp=None,
+  factor=(2,2,1)
+):
+  """
+  Downsamples an existing image layer that may be
+  sharded or unsharded to create a sharded layer.
+  
+  Only 2x2x1 downsamples are supported for now.
+  """
+  cv = downsample_scales.add_scale(
+    cloudpath, mip, 
+    preserve_chunk_size=True, chunk_size=chunk_size,
+    encoding=encoding, factor=factor
+  )
+  cv.mip = mip + 1
+  cv.scale["sharding"] = create_sharded_image_info(
+    dataset_size=cv.scale["size"], 
+    chunk_size=cv.scale["chunk_sizes"][0], 
+    encoding=cv.scale["encoding"], 
+    dtype=cv.dtype,
+    uncompressed_shard_bytesize=int(memory_target),
+  )
+  cv.commit_info()
+
+  shape = image_shard_shape_from_spec(
+    cv.scale["sharding"], cv.volume_size, cv.chunk_size
+  )
+  shape = Vec(*shape) * factor
+
+  cv.mip = mip
+  bounds = get_bounds(cv, None, mip, cv.meta.chunk_size(mip + 1))
+
+  class ImageShardDownsampleTaskIterator(FinelyDividedTaskIterator):
+    def task(self, shape, offset):
+      return partial(ImageShardDownsampleTask,
+        cloudpath,
+        shape=tuple(shape),
+        offset=tuple(offset),
+        mip=int(mip),
+        fill_missing=bool(fill_missing),
+        sparse=bool(sparse),
+        agglomerate=bool(agglomerate),
+        timestamp=timestamp,
+        factor=tuple(factor),
+      )
+
+    def on_finish(self):
+      job_details = {
+        "method": {
+          "task": "ImageShardDownsampleTask",
+          "cloudpath": cloudpath,
+          "shape": list(map(int, shape)),
+          "fill_missing": fill_missing,
+          "sparse": bool(sparse),
+          "bounds": [bounds.minpt.tolist(), bounds.maxpt.tolist()],
+          "mip": mip,
+          "agglomerate": agglomerate,
+          "timestamp": timestamp,
+        },
+        "by": operator_contact(),
+        "date": strftime("%Y-%m-%d %H:%M %Z"),
+      }
+
+      cv.provenance.sources = [cloudpath]
+      cv.provenance.processing.append(job_details)
+      cv.commit_provenance()
+
+  return ImageShardDownsampleTaskIterator(bounds, shape)
+
 def create_deletion_tasks(
     layer_path, mip=0, num_mips=5, 
     shape=None, bounds=None
@@ -534,7 +612,7 @@ def create_transfer_tasks(
     delete_black_uploads=False, background_color=0,
     agglomerate=False, timestamp=None, compress='gzip',
     factor=None, sparse=False, dest_voxel_offset=None,
-    memory_target=3.5e9, max_mips=5
+    memory_target=MEMORY_TARGET, max_mips=5
   ):
   """
   Transfer data to a new data layer. You can use this operation
@@ -962,7 +1040,7 @@ def create_quantize_tasks(
 
   class QuantizeTasksIterator(FinelyDividedTaskIterator):
     def task(self, shape, offset):
-      return QuantizeTask(
+      return partial(QuantizeTask,
         source_layer_path=src_layer,
         dest_layer_path=dest_layer,
         shape=shape.tolist(),
